package nrs.jester;

import java.io.*;
import java.awt.*;
import java.awt.event.*;
import java.lang.reflect.Method;
import java.math.BigDecimal;
import java.sql.*;
import java.text.*;
import java.util.*;
import javax.swing.*;

import nrs.jester.db.*;

/*
 * Java Stored Procedure Tester (Jester)
 * Copyright (C) 2002 Nick Sydenham <nsydenham@yahoo.co.uk>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA
 */

/**
 * Represents a connection to a database. It implements the Observable
 * interface so that Observers can be informed about connection/disconnection
 * events.
 *
 * @author Nicholas Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class DBConnection extends Observable implements Runnable
{
    // the connection to the database
    private Connection conn = null;
    // database factory call
    private DBCall call = null;
    
    // reference to app class
    private Jester jester = null;
    
    private final static String CONNECT_FILE = ".jester/connections.txt";
    private Map profiles = new TreeMap();
    private int dbType = 0;
    
    // used for Method reflection by the run method
    private Object[] methodArgs;
    private String methodName = null;

    /**
     * SPConnection constructor comment.
     */
    public DBConnection(Jester jt)
    {
	jester = jt;
	
	readConnections();
    }
    
    /**
     * Set the method and parameters used by the run method
     */
    void setRunParameters(String name, Object[] args) {
	methodName = name;
	methodArgs = args;
    }
	
    /**
     * Invoke the stored procedure
     * @param sp name of the sp
     */
    private void callSP(StoredProcedure sp, String[] values)
    {
	SimpleDateFormat datesdf = new SimpleDateFormat(jester.getDateFormat());
	SimpleDateFormat timesdf = new SimpleDateFormat(jester.getTimeFormat());
	SimpleDateFormat timestampsdf = new SimpleDateFormat(
	    jester.getDateFormat()+" "+jester.getTimeFormat());
	java.util.Date d;
	java.sql.Date sqlDate;
	Parameter param;
	String paramVal;
	int i;
	
	// prepare output types
	java.util.List op = sp.getOutParameters();
	int[] outputTypes = new int[op.size()];
	for (i=0; i<op.size(); i++)
	{
	    param = (Parameter)op.get(i);
	    outputTypes[i] = param.getDataType();
	}
	
	try
	{
	    call.prepareCall(conn, sp.getName(), sp.getInParameters().size(), outputTypes);
	    call.setQueryTimeout(jester.getTimeout());
	    
	    // set the input parameters
	    java.util.List ip = sp.getInParameters();
	    for (i=0; i<ip.size(); i++)
	    {
		param = (Parameter)ip.get(i);
		paramVal = values[i];
		
		switch (param.getDataType())
		{
		    case Types.CHAR:
		    case Types.VARCHAR:
		    case Types.LONGVARCHAR:
			call.setString(i+1, paramVal);
			break;
			
		    case Types.DATE:
			sqlDate = null;
			if (paramVal != null)
			{
			    try
			    {
				d = datesdf.parse(paramVal);
				sqlDate = new java.sql.Date(d.getTime());
			    } catch (ParseException pex) {
				jester.appendStatus("\nInvalid date:"+paramVal);
				return;
			    }
			}
			call.setDate(i+1, sqlDate);
			break;
			
		    case Types.BIGINT:
			if (paramVal != null)
			    call.setLong(i+1, Long.parseLong(paramVal));
			else
			    call.setNull(i+1, Types.BIGINT);
			break;

		    case Types.SMALLINT:
			if (paramVal != null)
			    call.setShort(i+1, Short.parseShort(paramVal));
			else
			    call.setNull(i+1, Types.SMALLINT);
			break;

		    case Types.TINYINT:
			if (paramVal != null)
			    call.setByte(i+1, Byte.parseByte(paramVal));
			else
			    call.setNull(i+1, Types.SMALLINT);
			break;

		    case Types.NUMERIC:
			if (paramVal != null)
			    call.setBigDecimal(i+1, BigDecimal.valueOf(Long.parseLong(paramVal)));
			else
			    call.setNull(i+1, Types.NUMERIC);
			break;
			
		    case Types.DECIMAL:
		    case Types.INTEGER:
			if (paramVal != null)
			    call.setInt(i+1, Integer.parseInt(paramVal));
			else
			    call.setNull(i+1, Types.INTEGER);
			break;
			
		    case Types.FLOAT:
			if (paramVal != null)
			    call.setFloat(i+1, Float.parseFloat(paramVal));
			else
			    call.setNull(i+1, Types.FLOAT);
			break;
			
		    case Types.DOUBLE:
			if (paramVal != null)
			    call.setDouble(i+1, Double.parseDouble(paramVal));
			else
			    call.setNull(i+1, Types.DOUBLE);
			break;
			
		    case Types.TIME:
			d = null;
			if (paramVal != null) {
			    try {
				d = timesdf.parse(paramVal);
			    } catch (ParseException pex) {
				jester.appendStatus("\nInvalid time: "+paramVal);
				return;
			    }
			}
			Time t = null;
			if (d != null) t = new Time(d.getTime());
			call.setTime(i+1, t);
			break;
			
		    case Types.TIMESTAMP:
			d = null;
			if (paramVal != null) {
			    try {
				d = timestampsdf.parse(paramVal);
			    } catch (ParseException pex) {
				jester.appendStatus("\nInvalid timestamp: "+paramVal);
				return;
			    }
			}
			Timestamp ts = null;
			if (d != null) ts = new Timestamp(d.getTime());
			call.setTimestamp(i+1, ts);
			break;
		}
		
		// copy value to default values map
		jester.addDefaultValue(param.getDescription(), paramVal);
	    }
	    
	    // execute the call and parse the data
	    ResultSet rs = call.executeQuery();
	    jester.setStatus("Called "+sp.getName());
	    if (rs.next()) {
		parseResultSet(rs);
		parseOutputParameters(sp);
	    } else {
		jester.setStatus("No data returned");
	    }
	    
	    if (rs != null) rs.close();
	} catch (SQLException sex) {
	    jester.printSQLException(sex);
	}
    }
    
    /**
     * Connect to the database. If the user is already connected, they will be disconnected.
     * Observers of this object will be informed.
     */
    void connect(ConnectionSetting cs)
    {
	try
	{
	    jester.setStatus("Loading driver and getting connection");
	    setDBType(cs.dbType);
	    call = CallFactory.createCall(cs.dbType);
	    conn = call.getConnection(cs, jester.getTimeout());
	    jester.setStatus("Connected");
	    
	    java.util.List colNames = new ArrayList();
	    colNames.add("Connected");
	    java.util.List rows = new ArrayList();
	    java.util.List row = new ArrayList();
	    row.add("No data");
	    rows.add(row);
	    jester.setTableData(colNames, rows);
	    
	    setChanged();
	    notifyObservers("C");
	} catch (SQLException sex) {
	    jester.appendStatus("\nCouldn't connect.\n");
	    jester.appendStatus(sex.toString());
	}
    }
    
    /**
     * Disconnect from the database. Observers of this object
     * will be informed.
     */
    void disconnect()
    {
	if (conn != null)
	{
	    try {
		conn.close();
	    } catch (SQLException sex) {}
	    conn = null;
	    jester.setStatus("Disconnected");
	} else {
	    jester.setStatus("Not connected");
	}
	
	setChanged();
	notifyObservers("D");
    }
    
    /**
     * Get metadata about the database
     */
    private void getMetaData(Integer options, String schema)
    {
	java.util.List rows = new ArrayList();
	java.util.List colNames = new ArrayList();
	java.util.List row;
	ResultSet rs;
	int opts = options.intValue();
	
	if (schema != null && schema.length() == 0) schema = null;
	
	colNames.add("Type");
	colNames.add("Name");
	colNames.add("Catalogue");
	colNames.add("Schema");
	colNames.add("Remarks");
	
	try
	{
	    DatabaseMetaData dbmd = conn.getMetaData();
	    
	    StringBuffer sb = new StringBuffer();
	    sb.append("Database: ");
	    sb.append(dbmd.getDatabaseProductName()+" "+dbmd.getDatabaseProductVersion());
	    sb.append("\nJDBC: ");
	    sb.append(dbmd.getDriverName()+" "+dbmd.getDriverVersion());
	    sb.append("\nMax connections: ");
	    sb.append(dbmd.getMaxConnections());
	    sb.append("\nSupports transactions: ");
	    sb.append(dbmd.supportsTransactions());
	    sb.append("\nRead-only: ");
	    sb.append(dbmd.isReadOnly());
	    sb.append("\nANSI91 entry level support: ");
	    sb.append(dbmd.supportsANSI92EntryLevelSQL());
	    sb.append("\nANSI92 intermediate support: ");
	    sb.append(dbmd.supportsANSI92IntermediateSQL());
	    sb.append("\nANSI92 full support: ");
	    sb.append(dbmd.supportsANSI92FullSQL());
	    jester.setMetadata(sb.toString());
	    
	    jester.setStatus("");
	    
	    // get tables
	    if ((opts & 2) == 2)
	    {
		String[] tableTypes = {"TABLE"};
		if (dbmd.supportsSchemasInTableDefinitions()) rs = dbmd.getTables(null, schema, null, tableTypes);
		else rs = dbmd.getTables(null, null, null, tableTypes);
		while (rs.next())
		{
		    row = new ArrayList();
		    row.add("Table");
		    row.add(rs.getString(3));
		    row.add(rs.getString(1));
		    row.add(rs.getString(2));
		    row.add(rs.getString(5));
		    rows.add(row);
		}
		rs.close();
		jester.appendStatus("Select a table to see its definition\n");
	    }
	    
	    // get stored procedures
	    if ((opts & 1) == 1)
	    {
		if (dbmd.supportsSchemasInProcedureCalls()) rs = dbmd.getProcedures(null, schema, null);
		else rs = dbmd.getProcedures(null, null, null);
		while (rs.next())
		{
		    row = new ArrayList();
		    row.add("Procedure");
		    row.add(rs.getString(3));
		    row.add(rs.getString(1));
		    row.add(rs.getString(2));
		    row.add(rs.getString(7));
		    rows.add(row);
		}
		rs.close();
		jester.appendStatus("Select a stored procedure to see its definition");
	    }
	    
	    jester.setTableData(colNames, rows);
	} catch (SQLException sex) {
	    jester.printSQLException(sex);
	    jester.finishedQuery();
	}
    }
    
    /**
     * Get the connection profiles
     * @return list of profiles
     */
    public Map getProfiles()
    {
	return profiles;
    }
    
    /**
     * Is there a connection to the database
     * @return true if there is a connection
     */
    public boolean isConnected()
    {
	if (conn != null) return true;
	else return false;
    }
    
    /**
     * Parse the ResultSet returned from the JDBC stored procedure.
     * <i>Note that this function expects only one row to be returned</i>
     * @param rs the ResultSet returned from the stored procedure call
     * @see #parseResultSet(ResultSet, Document, Node)
     */
    private void parseResultSet(ResultSet rs) throws SQLException
    {
	int i, columnCount, rowCount=0;
	Object sqlObj;
	String name, val;
	java.util.List rows, row;
	
	ResultSetMetaData md = rs.getMetaData();
	columnCount = md.getColumnCount();
	jester.appendStatus("\nGot "+columnCount+" columns");
	rows = new ArrayList(columnCount);
	
	java.util.List colNames = new ArrayList(columnCount);
	for (i=1; i<=columnCount; i++)
	{
	    colNames.add(md.getColumnName(i));
	}
	
	do
	{
	    row = new ArrayList(columnCount);
	    for (i=1; i<=md.getColumnCount(); i++)
	    {
		val = "";
		sqlObj = rs.getObject(i); // don't care about type
		if (sqlObj != null) val = sqlObj.toString().trim(); // check for null
		name = md.getColumnName(i).toLowerCase();
		row.add(val);
	    }
	    rows.add(row);
	    rowCount++;
	} while(rs.next());
	jester.appendStatus("\nGot "+rowCount+" rows");

	jester.setTableData(colNames, rows);
    }
    
    /**
     * See if any output parameters have been returned
     * @param sp the StoredProcedure being processed
     * @exception SQLException can't get output parameter
     */
    private void parseOutputParameters(StoredProcedure sp) throws SQLException
    {
	java.util.List inParams = sp.getInParameters();
	java.util.List outParams = sp.getOutParameters();
	Parameter out;
	int outCount = inParams.size() + 1;
	String desc;
	
	if (outParams.size() > 0) {
	    jester.appendStatus("\nOutput parameters");
	    
	    Iterator it = outParams.iterator();
	    while (it.hasNext()) {
		out = (Parameter)it.next();
		desc = out.getDescription();
		if (desc == null || desc.length() == 0) desc = "Out "+Integer.toString(outCount);
		jester.appendStatus("\n"+desc+" = ");
		
		Object val = call.getObject(outCount);
		if (val != null) jester.appendStatus(val.toString());
		outCount++;
	    }
	}
    }
    
    /**
     * Perform an SQL query
     * @param query the query text in SQL
     */
    private void performQuery(String query)
    {
	try
	{
	    Statement s = conn.createStatement();
	    if (call.supportsTimeout()) s.setQueryTimeout(jester.getTimeout());
	    ResultSet rs = s.executeQuery(query);
	    
	    jester.appendStatus("\nGot ResultSet");
	    if (rs.next())
		parseResultSet(rs);
	    else
		jester.setStatus("No data returned");
	    if (rs != null) rs.close();
	    s.close();
	} catch (SQLException sex) {
	    jester.printSQLException(sex);
	}
    }
    
    /**
     * Read the connections file if it exists
     */
    private void readConnections()
    {
	StringTokenizer st;
	String line, var, val;
	ConnectionSetting cs = null;
	String connName = null;
	
	// read in from the connections file if available
	try
	{
	    File home = new File(System.getProperty("user.home"));
	    File f = new File(home, CONNECT_FILE);
	    if (f.exists()) {
		BufferedReader reader = new BufferedReader(new FileReader(f));
		while ((line = reader.readLine()) != null)
		{
		    if (line.startsWith("["))
		    {
			// add previously defined profile
			if (cs != null) profiles.put(cs.profileName, cs);

			cs = new ConnectionSetting();
			connName = line.substring(1, line.length()-1);
			cs.profileName = connName;
			cs.driverProperties = new Properties();
			continue;
		    }

		    st = new StringTokenizer(line, "=");
		    if (st.countTokens() == 2)
		    {
			if (cs == null)
			{
			    // invalid file
			    System.out.println("Invalid connections file "+CONNECT_FILE);
			    return;
			}
			var = st.nextToken().trim();
			val = st.nextToken().trim();
			if (var.equalsIgnoreCase("host")) {
			    cs.host = val;
			} else if (var.equalsIgnoreCase("dbName")) {
			    cs.dbName = val;
			} else if (var.equalsIgnoreCase("dbType")) {
			    try {
				dbType = Integer.parseInt(val);
				cs.dbType = dbType;
			    } catch (NumberFormatException nfex) {}
			} else if (var.equalsIgnoreCase("user")) {
			    cs.user = val;
			} else if (var.equalsIgnoreCase("password")) {
			    cs.password = val;
			    cs.savePassword = true;
			} else if (var.equalsIgnoreCase("spDefFile")) {
			    cs.spDefFile = val;
			} else {
			    cs.driverProperties.setProperty(var, val);
			}
		    }
		}
		
		// add last profile
		if (cs != null) profiles.put(cs.profileName, cs);
	    }
        } catch (Exception ex) {
	    System.out.println("Error reading connections: "+ex.getMessage());
	}
    }
    
    /**
     * Save the connections specified in the Connect dialog
     * to a file so they can be reused next time
     */
    void saveConnections()
    {
	ConnectionSetting cs;
	
	try
	{
	    File home = new File(System.getProperty("user.home"));
	    File f = new File(home, CONNECT_FILE);

	    // delete existing file
	    if (f.exists())
	    {
		f.delete();
	    }
	    
	    // write all data to file
	    PrintWriter writer = new PrintWriter(new BufferedWriter(new FileWriter(f)));
	    
	    // write out hosts
	    Iterator it = profiles.values().iterator();
	    while (it.hasNext())
	    {
		cs = (ConnectionSetting)it.next();
		writer.println("["+cs.profileName+"]");
		writer.println(cs.toString());
	    }
	    
	    writer.close();
	} catch (Exception ex) {}
    }
    
    /**
     * Set up the connection properties based on the database selected
     * @param db database selection
     */
    private void setDBType(int db)
    {
	dbType = db;
    }
    
    /**
     * Show details about a stored procedure or table
     * @param type either Procedure or Table
     * @param schema name of the schema to search in
     * @param name the name of the procedure or table
     */
    private void showDetails(String type, String schema, String name)
    {
	DatabaseMetaData dbmd;
	ResultSet rs;
	java.util.List rows = new ArrayList();
	java.util.List colNames = new ArrayList();
	java.util.List rowList;
	short colType;
	String colT, tmpSchema=null, colName;

	try
	{
	    if (type.equals("Procedure"))
	    {
		dbmd = conn.getMetaData();
		
		if (dbmd.supportsSchemasInProcedureCalls()) tmpSchema = schema;
		rs = dbmd.getProcedureColumns(null, tmpSchema, name, null);
		while (rs.next())
		{
		    rowList = new ArrayList();
		    rowList.add(name);
		    rowList.add(rs.getString(4));
		    colType = rs.getShort(5);
		    switch (colType)
		    {
			case DatabaseMetaData.procedureColumnIn:
			    colT = "IN";
			    break;
			    
			case DatabaseMetaData.procedureColumnInOut:
			    colT = "IN/OUT";
			    break;
			    
			case DatabaseMetaData.procedureColumnOut:
			    colT = "OUT";
			    break;
			   
			case DatabaseMetaData.procedureColumnReturn:
			    colT = "Return";
			    break;
			    
			case DatabaseMetaData.procedureColumnResult:
			    colT = "Result";
			    break;
			    
			default:
			    colT = "?";
		    }
		    rowList.add(colT);
		    
		    rowList.add(rs.getString(7));
		    rowList.add(rs.getObject(9).toString());
		    rowList.add(rs.getObject(8).toString());
		    rowList.add(rs.getString(13));
		    rows.add(rowList);
		}
		
		colNames.add("Procedure");
		colNames.add("Column name");
		colNames.add("Column type");
		colNames.add("Data type");
		colNames.add("Length");
		colNames.add("Precision");
		colNames.add("Remarks");
		
		jester.setTableData(colNames, rows);
	    }
	    else if (type.equals("Table"))
	    {
		dbmd = conn.getMetaData();
		
		if (dbmd.supportsSchemasInTableDefinitions()) tmpSchema = schema;
		
		// get primary keys of table
		rs = dbmd.getPrimaryKeys(null, tmpSchema, name);
		java.util.List primaryKeys = new ArrayList();
		while (rs.next()) {
		    colName = (String)rs.getString(4);
		    primaryKeys.add(colName);
		}
		
		rs = dbmd.getColumns(null, tmpSchema, name, null);
		while (rs.next())
		{
		    rowList = new ArrayList();
		    rowList.add(name);
		    colName = rs.getString(4);
		    if (primaryKeys.contains(colName)) colName += "*";
		    rowList.add(colName);
		    rowList.add(rs.getString(6));
		    rowList.add(rs.getObject(7).toString());
		    rowList.add(rs.getString(13));
		    rowList.add(rs.getString(18));
		    rowList.add(rs.getString(12));
		    rows.add(rowList);
		}
		
		colNames.add("Table");
		colNames.add("Column");
		colNames.add("Type");
		colNames.add("Size");
		colNames.add("Default");
		colNames.add("Nullable");
		colNames.add("Remarks");
		
		jester.setTableData(colNames, rows);
	    }
	} catch (SQLException sex) {
	    jester.printSQLException(sex);
	}
    }
    
    /**
     * Get a list of stored procedures from the database
     * using meta-data.
     * @param spReader the list model
     * @param schema name of the schema to search in
     */
    private void getStoredProcedures(SPReader spReader, String schema)
    {
	StoredProcedure sp;
	Parameter param = null;
	String spName, colName, colDesc, typeName;
        Short dataType;
	short colType, colDataType;
	int colLength;
	boolean nullable;
	java.util.List sps = new ArrayList();
	ResultSet prs, crs, trs;
	
	try {	    
	    // set up type names for this database
	    Map typeNames = new HashMap();
	    DatabaseMetaData dbmd = conn.getMetaData();
	    trs = dbmd.getTypeInfo();
	    while (trs.next()) {
		dataType = new Short(trs.getShort(2));
		typeName = trs.getString(1);
		typeNames.put(dataType, typeName);
	    }
	    
	    // get stored procedures
	    if (!dbmd.supportsSchemasInProcedureCalls()) schema = null;
	    prs = dbmd.getProcedures(null, schema, null);
	    while (prs.next()) {
		sp = new StoredProcedure(prs.getString(3));
		sp.setDescription(prs.getString(7));
		sps.add(sp);
	    }

	    // get stored procedure column definitions
	    Iterator it = sps.iterator();
	    while (it.hasNext()) {
		sp = (StoredProcedure)it.next();
		crs = dbmd.getProcedureColumns(null, schema, sp.getName(), null);
		while (crs.next()) {
		    colName = crs.getString(4);
		    colType = crs.getShort(5);
		    colDataType = (short)crs.getInt(6);
		    colLength = crs.getInt(9);
		    colDesc = crs.getString(13);
		    if (colDesc == null || colDesc.length() == 0) {
			colDesc = (String)typeNames.get(new Short(colDataType));
		    }
		    if (crs.getShort(12) == DatabaseMetaData.procedureNullable)
			nullable = true;
		    else
			nullable = false;
		    
		    if (colType == dbmd.procedureColumnIn) {
			param = sp.addInParameter(colDataType, null, colDesc);
			param.setLength(colLength);
			param.setNullable(nullable);
		    } else if (colType == dbmd.procedureColumnOut) {
			param = sp.addOutParameter(colDataType, crs.getString(13));
			param.setLength(colLength);
			param.setNullable(nullable);
		    }
		}
	    }
	    
	    spReader.setStoredProcedures(sps);
	    
	} catch (SQLException sex) {
	    jester.printSQLException(sex);
	    jester.finishedQuery();
	}
    }
    
    /**
     * Execute in a separate thread so that long actions
     * can be aborted by the GUI.
     */
    public void run()
    {
	Object methodArg;
	
	Class[] methodFindArgs = new Class[methodArgs.length];
	
	for (int i=0; i<methodArgs.length; i++)
	{
	    methodArg = methodArgs[i];
	    if (methodArg == null) methodFindArgs[i] = (new String()).getClass();
	    else methodFindArgs[i] = methodArg.getClass();
	}
	
	try {
	    Method runMethod = getClass().getDeclaredMethod(methodName, methodFindArgs);
	    runMethod.invoke(this, methodArgs);
	} catch (Exception ex) {
	    // real bad - shouldn't happen
	    jester.setStatus("Couldn't call method "+methodName);
	    jester.appendStatus("\n"+ex.getMessage());
	}
	
	// tell Jester we've finished
	jester.finishedQuery();
    }
}
