package nrs.jester.db;

import java.math.BigDecimal;
import java.sql.*;
import java.util.*;

import nrs.jester.ConnectionSetting;

/*
 * Java Stored Procedure Tester (Jester)
 * Copyright (C) 2002 Nick Sydenham <nsydenham@yahoo.co.uk>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA
 */

/**
 * An abstraction of the CallableStatment class which allows concrete implementations
 * to do database specific operations. In particular, these are setting up stored
 * procedure calls and obtaining result sets.
 *
 * @author Nicholas Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public abstract class DBCall
{
    protected CallableStatement call = null;

    /**
     * Close the CallableStatement
     */
    public void close() throws SQLException {
        if (call != null) call.close();
        call = null;
    }
    
    /**
     * Execute the call
     * @return true if the call executed successfully
     */
    public boolean execute() throws SQLException {
        return call.execute();
    }
    /**
     * Executes a SQL statement that returns a single ResultSet
     * @return a ResultSet that contains the data produced by the query; never null
     * @exception java.sql.SQLException if a database access error occurs
     */
    public ResultSet executeQuery() throws SQLException {
        return call.executeQuery();
    }
    
    /**
     * Close the connection
     */
    protected void finalize() {
        try {
            close();
        } catch (SQLException sex) {}
    }
    
    /**
     * Get a Connection to the database
     * @param cs ConnectionSetting
     * @return an SQL connection
     * @exception java.sql.SQLException if a connection can't be made
     */
    public abstract Connection getConnection(ConnectionSetting cs, int timeout) throws SQLException;
    
    public BigDecimal getBigDecimal(int param) throws SQLException {
	return call.getBigDecimal(param);
    }
    
    public byte getByte(int param) throws SQLException {
	return call.getByte(param);
    }
    
    public double getDouble(int param) throws SQLException {
	return call.getDouble(param);
    }
    
    public float getFloat(int param) throws SQLException {
	return call.getFloat(param);
    }
    
    /**
     * Get an integer value
     * @param param column index
     * @return int
     */
    public int getInt(int param) throws SQLException {
        return call.getInt(param);
    }

    public long getLong(int param) throws SQLException {
	return call.getLong(param);
    }
    
    public Object getObject(int param) throws SQLException {
	return call.getObject(param);
    }
    
    public short getShort(int param) throws SQLException {
	return call.getShort(param);
    }
    
    /**
     * Get a String from a column
     * @param param the column index
     * @return java.lang.String
     */
    public String getString(int param) throws SQLException {
        return call.getString(param);
    }
    
    public Time getTime(int param) throws SQLException {
	return call.getTime(param);
    }
    
    public Timestamp getTimestamp(int param) throws SQLException {
	return call.getTimestamp(param);
    }
    
    /**
     * Prepare a stored procedure call
     * @param conn the connection to the database
     * @param procedure the stored procedure name including whatever package it's in
     * @param inputParams the number of input parameters
     * @param outputTypes the output types
     * @exception SQLException if it goes wrong
     */
    public abstract void prepareCall(Connection conn, String procedure, int inputParams, int[] outputTypes) throws SQLException;

    /**
     * Get a ResultSet from the call
     * @return java.sql.ResultSet
     */
    public abstract ResultSet getResultSet() throws SQLException;

    /**
     * Register an output parameter
     */
    public void registerOutParameter(int param, int type) throws SQLException {
        call.registerOutParameter(param, type);
    }

    public void setBigDecimal(int param, BigDecimal val) throws SQLException {
	call.setBigDecimal(param, val);
    }
    
    public void setByte(int param, byte val) throws SQLException {
	call.setByte(param, val);
    }
    
    /**
     * Set a date parameter
     * @param param parameter number
     * @param date date to set
     * @exception java.sql.SQLException The exception description.
     */
    public void setDate(int param, java.sql.Date date) throws SQLException {
        if (date != null) {
            call.setDate(param, date);
        } else {
            call.setNull(param, Types.DATE);
        }
    }
    
    /**
     * Set an integer parameter
     * @param param column index
     * @param val the value
     */
    public void setInt(int param, int val) throws SQLException {
        call.setInt(param, val);
    }
    
    public void setLong(int param, long val) throws SQLException {
	call.setLong(param, val);
    }
    
    /**
     * Set an SQL null parameter
     * @param paramIndex parameter index
     * @param sqlType SQL type from java.sql.Types
     */
    public void setNull(int paramIndex, int sqlType) throws SQLException {
        call.setNull(paramIndex, sqlType);
    }
    
    /**
     * Set the call timeout in seconds
     * @param timeout int
     */
    public void setQueryTimeout(int timeout) throws SQLException {
        call.setQueryTimeout(timeout);
    }
    
    public void setShort(int param, short val) throws SQLException {
	call.setShort(param, val);
    }
    
    /**
     * Set a string in a stored procedure call
     * @param param column index
     * @param val value
     */
    public void setString(int param, String val) throws SQLException {
        if (val == null) call.setNull(param, Types.CHAR);
        else call.setString(param, val);
    }
    
    /**
     * Set a float in a stored procedure call
     * @param param column index
     * @param val value
     */
    public void setFloat(int param, float val) throws SQLException {
	call.setFloat(param, val);
    }
    
    /** Set a double in a stored procedure call */
    public void setDouble(int param, double val) throws SQLException {
	call.setDouble(param, val);
    }
    
    /** Set a time in a stored procedure call */
    public void setTime(int param, Time val) throws SQLException {
	call.setTime(param, val);
    }
    
    /** Set a timestamp in a stored procedure call */
    public void setTimestamp(int param, Timestamp val) throws SQLException {
	call.setTimestamp(param, val);
    }
    
    /**
     * Get a reference to the CallableStatement
     * @return callable statement
     */
    protected CallableStatement spcall() {
        return call;
    }
    
    /**
     * Get optional connection properties
     * @return an array of connection property names
     */
    public abstract String[] getConnectionProperties();
    
    /**
     * Set up the connection
     * @param url the database connection URL
     * @param cs ConnectionSetting
     * @param timeout connection timeout
     * @exception SQLException couldn't connect
     */
    protected Connection makeConnection(String url, ConnectionSetting cs, int timeout) throws SQLException
    {
	Properties props = new Properties();
	if (cs.driverProperties != null) {
	    Enumeration enum = cs.driverProperties.keys();
	    String key;
	    while (enum.hasMoreElements()) {
		key = (String)enum.nextElement();
		props.put(key, cs.driverProperties.getProperty(key));
	    }
	}
        if (cs.user != null) props.put("user", cs.user);
        if (cs.password != null) props.put("password", cs.password);
        
	if (timeout > 0) DriverManager.setLoginTimeout(timeout);
	
        return DriverManager.getConnection(url, props);
    }

    /**
     * Return true if the DB supports query timeouts
     * @return true if query timeout is supported
     */
    public boolean supportsTimeout()
    {
	return true;
    }
}
