package nrs.jester.db;

import java.sql.*;
import java.util.Properties;
import oracle.jdbc.driver.*;

import nrs.jester.ConnectionSetting;

/*
 * Java Stored Procedure Tester (Jester)
 * Copyright (C) 2002 Nick Sydenham <nsydenham@yahoo.co.uk>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA
 */

/**
 * Oracle specific instance of the DBCall class.
 *
 * @author Nicholas Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public abstract class OracleBaseCall extends DBCall
{
    private final static String DRIVER = "oracle.jdbc.driver.OracleDriver";
    protected String _URL = null;
    private final static String[] connProps = {
	"defaultBatchValue", "defaultRowPrefetch", "includeSynonyms",
	"internal_logon", "remarksReporting"
    };
    
    /**
     * Get a Connection to an Oracle DB
     * @param cs ConnectionSetting
     * @return java.sql.Connection
     * @exception java.sql.SQLException The exception description.
     */
    public Connection getConnection(ConnectionSetting cs, int timeout) throws SQLException
    {
	try {
	    Class.forName(DRIVER);
	} catch (ClassNotFoundException cnfex) {
	    throw new SQLException("Oracle JDBC driver not in classpath");
	}
	
	StringBuffer sb = new StringBuffer(_URL);
	sb.append(":@");
	sb.append(cs.host);
	sb.append(":");
	if (cs.port > 0) {
	    sb.append(cs.port);
	    sb.append(":");
	}
	sb.append(cs.dbName);
	
        return makeConnection(sb.toString(), cs, timeout);
    }
    
    /**
     * Get the ResultSet returned from a stored procedure call
     */
    public ResultSet getResultSet() throws SQLException
    {    
        return (OracleResultSet)(call.getObject(1));
    }
    
    /**
     * Prepare a stored procedure call. As this is Oracle it will expect the SP to
     * return a stored procedure which is allocated to output parameter 1.
     * @param conn the connection to the database
     * @param procedure the stored procedure name including whatever package it's in
     * @param inputParams the number of input parameters
     * @param outputTypes the output types
     * @exception SQLException if it goes wrong
     */
    public void prepareCall(Connection conn, String procedure, int inputParams, int[] outputTypes) throws SQLException
    {
        int i;
        
        StringBuffer sql = new StringBuffer("{? = call ");
        sql.append(procedure);
        sql.append("(");
        
        for (i=0; i<inputParams; i++) {
            sql.append("?,");
        }
        
        for (i=0; i<outputTypes.length; i++) {
            sql.append("?,");
        }
        
        if (inputParams>0 || outputTypes.length>0)
            sql.setLength(sql.length()-1); // remove the last comma
        
        sql.append(")}");
        
        call = conn.prepareCall(sql.toString());
        call.registerOutParameter(1, OracleTypes.CURSOR);
        
        // register the output parameters
        for (i=0; i<outputTypes.length; i++) {
            call.registerOutParameter(inputParams+i+2, outputTypes[i]);
        }
    }
    
    /**
     * Set an integer parameter
     * @param param column index
     * @param val int
     */
    public void setInt(int param, int val) throws SQLException {
        // the first parameter is the returned ResultSet
        call.setInt(param+1, val);
    }
    
    /**
     * Set a string parameter
     * @param param column index
     * @param val java.lang.String
     */
    public void setString(int param, String val) throws SQLException {
        // the first parameter is the returned ResultSet
        call.setString(param+1, val);
    }
    
    /**
     * Get optional connection properties
     * @return an array of connection property names
     */
    public String[] getConnectionProperties()
    {
	return connProps;
    }    
}

